from .env import env_var, env_switch
from .db import table_exists, query_db, get_max_value, get_min_value
from .time import previous_date, previous_hour
from datetime import datetime, timedelta


default_functions = {
    "datetime": datetime,
    "env_switch": env_switch,
    "env_var": env_var,
    "get_max_value": get_max_value,
    "get_min_value": get_min_value,
    "previous_date": previous_date,
    "previous_hour": previous_hour,
    "table_exists": table_exists,
    "timedelta": timedelta,
    "query_db": query_db,
}
default_constructors = {
    "!" + k: v
    for k, v in default_functions.items()
    if k
    in (
        "datetime",
        "env_switch",
        "env_var",
        "previous_date",
        "previous_hour",
        "timedelta",
    )
}



from absql.files import accepted_file_types
from absql.files.loader import generate_loader
from absql.render import render_text, render_context, render_file


class Runner:
    def __init__(
        self,
        extra_constructors=None,
        replace_only=False,
        file_context_from=None,
        partial_kwargs=None,
        **extra_context,
    ):
        self.extra_context = dict(extra_context)
        self.loader = generate_loader(extra_constructors or [])
        self.replace_only = replace_only
        self.file_context_from = file_context_from
        self.partial_kwargs = partial_kwargs or ["engine"]

    @staticmethod
    def render_text(
        text, replace_only=False, pretty_encode=False, partial_kwargs=None, **vars
    ):
        return render_text(
            text=text,
            replace_only=replace_only,
            pretty_encode=pretty_encode,
            partial_kwargs=partial_kwargs,
            **vars,
        )

    @staticmethod
    def render_context(extra_context=None, file_contents=None, partial_kwargs=None):
        return render_context(
            extra_context=extra_context,
            file_contents=file_contents,
            partial_kwargs=partial_kwargs,
        )

    @staticmethod
    def render_file(
        file_path,
        loader=None,
        replace_only=False,
        extra_constructors=None,
        file_context_from=None,
        pretty_encode=False,
        partial_kwargs=None,
        **extra_context,
    ):
        return render_file(
            file_path=file_path,
            loader=loader,
            replace_only=replace_only,
            extra_constructors=extra_constructors,
            file_context_from=file_context_from,
            pretty_encode=pretty_encode,
            partial_kwargs=partial_kwargs,
            **extra_context,
        )

    def render(self, text, pretty_encode=False, replace_only=None, **extra_context):
        """
        Given text or a file path, render SQL with the a combination of
        the vars in the file and any extras passed to extra_context during
        the instantiation of the runner.
        """

        current_context = self.extra_context.copy()
        current_context.update(extra_context)

        if text.endswith(accepted_file_types):
            rendered = render_file(
                file_path=text,
                loader=self.loader,
                replace_only=replace_only or self.replace_only,
                file_context_from=self.file_context_from,
                pretty_encode=pretty_encode,
                partial_kwargs=self.partial_kwargs,
                **current_context,
            )
        else:
            rendered = render_text(
                text=text,
                replace_only=replace_only or self.replace_only,
                pretty_encode=pretty_encode,
                partial_kwargs=self.partial_kwargs,
                **render_context(current_context, partial_kwargs=self.partial_kwargs),
            )
        return rendered

    def set_context(self, **context):
        self.extra_context.update(context)



import yaml
from absql.functions import default_constructors


def scalar_to_value(scalar, constructor_dict):
    """
    Converts a YAML ScalarNode to its underlying Python value
    """
    type = scalar.tag.split(":")[-1]
    val = scalar.value
    if isinstance(scalar, yaml.MappingNode):
        val = node_converter(scalar, constructor_dict)
        func = constructor_dict.get(type)
        return func(**val)
    if isinstance(scalar, yaml.SequenceNode):
        val = node_converter(scalar, constructor_dict)
        func = constructor_dict.get(type)
        return func(*val)
    if type.startswith("!"):
        func = constructor_dict.get(type)
        return func(val)
    # Handle null type - https://yaml.org/type/null.html
    if type == "null":
        return None
    return eval('{type}("""{val}""")'.format(type=type, val=val))


def node_converter(x, constructor_dict):
    """
    Converts YAML nodes of varying types into Python values,
    lists, and dictionaries
    """
    if isinstance(x, yaml.ScalarNode):
        # "I am an atomic value"
        return yaml.load(x.value, yaml.SafeLoader)
    if isinstance(x, yaml.SequenceNode):
        # "I am a list"
        return [scalar_to_value(v, constructor_dict) for v in x.value]
    if isinstance(x, yaml.MappingNode):
        # "I am a dict"
        return {
            scalar_to_value(v[0], constructor_dict): scalar_to_value(
                v[1], constructor_dict
            )
            for v in x.value
        }


def wrap_yaml(func, constructor_dict):
    """Turn a function into one that can be run on a YAML input"""

    def ret(loader, x):
        value = node_converter(x, constructor_dict)

        if value is not None:

            if isinstance(value, list):
                return func(*value)

            if isinstance(value, dict):
                return func(**value)

            return func(value)

        else:
            return func()

    return ret


def generate_loader(extra_constructors=None):
    """Generates a SafeLoader with both default and custom constructors"""
    loader = yaml.SafeLoader
    extra_constructors = extra_constructors or []
    default_constructor_dict = default_constructors.copy()

    if isinstance(extra_constructors, list) and len(extra_constructors) > 0:
        extra_constructors = {
            ("!" + func.__name__): func for func in extra_constructors
        }

    if len(extra_constructors) > 0:
        default_constructor_dict.update(extra_constructors)
    for tag, func in default_constructor_dict.items():
        loader.add_constructor(tag, wrap_yaml(func, default_constructor_dict))
    return loader



from inspect import signature
from functools import partial


def nested_apply(x, f):
    if x.__class__.__name__ in ["dict", "list", "str"]:
        if isinstance(x, str):
            x = f(x)
        elif isinstance(x, list):
            x = [nested_apply(i, f) for i in x]
        elif isinstance(x, dict):
            for k in x:
                x[k] = nested_apply(x[k], f)
    else:
        return x
    return x


def get_function_arg_names(func):
    return list(signature(func).parameters.keys())


def partialize_function(func, partial_kwargs=None, **kwargs):
    partial_kwargs = partial_kwargs or ["engine"]
    function_args = get_function_arg_names(func)

    kwargs_to_partialize = {
        k: v for k, v in kwargs.items() if k in function_args and k in partial_kwargs
    }

    if kwargs_to_partialize:
        return partial(func, **kwargs_to_partialize)
    else:
        return func



from absql.utils import nested_apply, get_function_arg_names, partialize_function


def test_simple_apply():
    d = {"a": "a", "b": "b", "c": 5}
    got = nested_apply(d, lambda x: x + "zzz")
    want = {"a": "azzz", "b": "bzzz", "c": 5}
    assert got == want


def test_nested_list():
    d = {"a": "a", "b": "b", "c": [5, "d"]}
    got = nested_apply(d, lambda x: x + "zzz")
    want = {"a": "azzz", "b": "bzzz", "c": [5, "dzzz"]}
    assert got == want


def test_nested_dict():
    d = {"a": "a", "b": "b", "c": {5: "d"}}
    got = nested_apply(d, lambda x: x + "zzz")
    want = {"a": "azzz", "b": "bzzz", "c": {5: "dzzz"}}
    assert got == want


def test_get_function_arg_names():
    def func(a, b):
        return a + b

    got = get_function_arg_names(func)
    want = ["a", "b"]
    assert got == want


def test_partialize_function():
    def simple_func(a, engine):
        return a + engine

    simple_func = partialize_function(simple_func, engine=7)
    got = simple_func(3)
    want = 10
    assert got == want



from inspect import cleandoc
from jinja2 import Environment, DebugUndefined
from absql.text import (
    clean_spacing,
    create_replacements,
    flatten_inputs,
    pretty_encode_sql,
)
from absql.files import parse
from absql.files.loader import generate_loader
from absql.functions import default_functions
from absql.utils import nested_apply, partialize_function


def render_text(
    text, replace_only=False, pretty_encode=False, partial_kwargs=None, **vars
):
    """
    Given some text, render the template with the vars.
    If a templated variable is unknown, leave it alone.
    """
    if replace_only:
        text = clean_spacing(text)
        flat_vars = flatten_inputs(**vars)
        replacements = create_replacements(**flat_vars)
        for k, v in replacements.items():
            text = text.replace(k, str(v))
        text = cleandoc(text)

    else:
        env = Environment(undefined=DebugUndefined)

        for k, v in vars.items():
            if v.__class__.__name__ == "function":
                vars[k] = partialize_function(v, partial_kwargs=partial_kwargs, **vars)
                env.filters[k] = vars[k]

        template = env.from_string(text)
        text = cleandoc(template.render(**vars))

    if pretty_encode:
        return pretty_encode_sql(text)
    else:
        return text


def render_context(extra_context=None, file_contents=None, partial_kwargs=None):
    """
    Render context dictionaries passed through a function call or
    file frontmatter (file_contents), with file_contents taking
    precedence over other all other provided context.
    """
    rendered_context = default_functions.copy()
    if extra_context:
        rendered_context.update(**extra_context)
    if file_contents:
        rendered_context.update(**file_contents)
    rendered_context = nested_apply(
        rendered_context,
        lambda x: render_text(x, partial_kwargs=partial_kwargs, **rendered_context),
    )
    return rendered_context


def render_file(
    file_path,
    loader=None,
    replace_only=False,
    extra_constructors=None,
    file_context_from=None,
    pretty_encode=False,
    partial_kwargs=None,
    **extra_context,
):
    """
    Given a file path, render SQL with a combination of
    the vars in the file and any extras passed to extra_context.
    """
    if loader is None:
        loader = generate_loader(extra_constructors or [])

    file_contents = parse(file_path, loader=loader)

    sql = file_contents["sql"]
    file_contents.pop("sql")

    if file_context_from:
        file_contents.update(file_contents.get(file_context_from, {}))
        file_contents.pop(file_context_from, {})

    rendered_context = render_context(extra_context, file_contents, partial_kwargs)

    rendered = render_text(
        text=sql,
        replace_only=replace_only,
        pretty_encode=pretty_encode,
        partial_kwargs=partial_kwargs,
        **rendered_context,
    )

    return rendered



import re
from flatdict import FlatDict
from sql_metadata import Parser
from colorama import Fore


def clean_spacing(text):
    text = re.sub("\\{\\{", "{{ ", text)
    text = re.sub("\\}\\}", " }}", text)
    text = re.sub("\\{\\{\\s+", "{{ ", text)
    text = re.sub("\\s+\\}\\}", " }}", text)
    return text


def create_replacements(**kwargs):
    replacements = {}
    for k, v in kwargs.items():
        replacement = {"{{{{{k}}}}}".format(k=k): v, "{{{{ {k} }}}}".format(k=k): v}
        replacements.update(replacement)
    return replacements


def flatten_inputs(**kwargs):
    flattened = FlatDict(kwargs, delimiter=".")
    return flattened


def pretty_encode_sql(
    query, keyword_color=Fore.LIGHTCYAN_EX, quote_color=Fore.LIGHTMAGENTA_EX
):
    p = Parser(query)

    keywords = list(set([t.value for t in p.tokens if t.is_keyword]))

    bold_start = "\033[1m"
    bold_end = "\033[0m"

    replacements = {k: bold_start + keyword_color + k + bold_end for k in keywords}

    for keyword, formatted_keyword in replacements.items():
        query = re.sub(r"\b{k}\b".format(k=keyword), formatted_keyword, query)

    quotes = [
        # single_quotes
        "'{t}'".format(t=text)
        for text in list(set(re.findall("'([^']*)'", query)))
    ] + [
        # double quotes
        '"{t}"'.format(t=text)
        for text in list(set(re.findall('"([^"]*)"', query)))
    ]

    replacements = {q: quote_color + q + Fore.RESET for q in quotes}

    for quote, formatted_quote in replacements.items():
        query = re.sub(quote, formatted_quote, query)

    return query



import pytest
import mock
import os
from absql import Runner


@pytest.fixture(autouse=True)
def mock_settings_env_vars():
    with mock.patch.dict(os.environ, {"name": "Bob"}):
        yield


@pytest.fixture()
def runner():
    return Runner(greeting="Hello")


@pytest.fixture()
def contextless_runner():
    return Runner()


@pytest.fixture
def simple_sql_path():
    return "tests/files/simple.sql"


@pytest.fixture
def no_frontmatter_path():
    return "tests/files/no_frontmatter.sql"


def test_render_simple_sql(runner, simple_sql_path):
    sql = runner.render(simple_sql_path)
    assert sql == "SELECT * FROM my_table"


def test_render_no_frontmatter(runner, no_frontmatter_path):
    sql = runner.render(no_frontmatter_path)
    assert sql == "SELECT * FROM Hello"


def test_render_text_only(runner):
    got = runner.render("{{greeting}}, {{env_var('name')}}!")
    want = "Hello, Bob!"
    assert got == want


def test_render_extra_context(runner):
    got = runner.render("{{new_greeting}}, {{env_var('name')}}!", new_greeting="Hey")
    want = "Hey, Bob!"
    assert got == want


def test_contextless_runner(contextless_runner):
    got = contextless_runner.render("{{no_greeting}}, Bill!")
    want = "{{ no_greeting }}, Bill!"
    assert got == want


def test_replace_only_changes(runner):
    original = runner.render("{{env_switch(foo='address')}} and {{greeting}}")
    assert original == "value_unspecified and Hello"

    replaced_only = runner.render(
        "{{env_switch(foo='address')}} and {{greeting}}", replace_only=True
    )
    assert replaced_only == "{{ env_switch(foo='address') }} and Hello"

    original_2 = runner.render("{{env_switch(foo='address')}} and {{greeting}}")
    assert original_2 == "value_unspecified and Hello"



from absql.functions.env import env_var
from sqlalchemy import create_engine, text
from sqlalchemy.engine.base import Engine


def table_exists(table_location, engine_env="AB__URI", engine=None):
    table_parts = split_parts(table_location)
    engine = (
        handle_engine(env_var(engine_env)) if engine is None else handle_engine(engine)
    )
    if hasattr(engine, "has_table"):
        return engine.has_table(
            table_name=table_parts["target"], schema=table_parts["namespace"]
        )
    else:
        return engine.reflection.Inspector.has_table(
            table_name=table_parts["target"], schema=table_parts["namespace"]
        )


def query_db(query, engine_env="AB__URI", engine=None):
    engine = (
        handle_engine(env_var(engine_env)) if engine is None else handle_engine(engine)
    )
    with engine.connect() as connection:
        return connection.execute(text(query)).fetchall()


def handle_engine(engine):
    if isinstance(engine, Engine):
        return engine
    else:
        return create_engine(engine)


def get_max_value(field_location, engine_env="AB__URI", engine=None):
    field_parts = split_parts(field_location)

    query = "SELECT MAX({field}) AS value FROM {table}".format(
        field=field_parts["target"], table=field_parts["namespace"]
    )
    try:
        return query_db(query, engine_env, engine)[0].value
    except Exception:
        return None


def get_min_value(field_location, engine_env="AB__URI", engine=None):
    field_parts = split_parts(field_location)

    query = "SELECT MIN({field}) AS value FROM {table}".format(
        field=field_parts["target"], table=field_parts["namespace"]
    )
    try:
        return query_db(query, engine_env, engine)[0].value
    except Exception:
        return None


def split_parts(location):
    parts = {}
    location_parts = location.split(".")
    target = location_parts[-1]
    namespace = (
        None
        if len(location_parts) == 1
        else ".".join(location_parts[: len(location_parts) - 1])
    )
    parts["target"] = target
    parts["namespace"] = namespace
    return parts



import pytest
from absql import Runner


@pytest.fixture()
def planet():
    def planet(planet_name):
        return planet_name

    return planet


def test_render_text_extra_partial(planet):

    got = Runner.render_text(
        "Hello {{planet()}}",
        planet=planet,
        planet_name="Earth",
        partial_kwargs=["planet_name"],
    )
    want = "Hello Earth"
    assert got == want


def test_default_render_text_extra_partial_fails(planet):
    with pytest.raises(TypeError):
        Runner.render_text("Hello {{planet()}}", planet=planet, planet_name="Earth")


def test_render_file_extra_partial(planet):
    got = Runner.render_file(
        "tests/files/partial_planet.sql",
        planet=planet,
        planet_name="Earth",
        partial_kwargs=["planet_name"],
    )
    want = "Hello Earth, Goodbye Earth"
    assert got == want


def test_render_runner_extra_partial(planet):
    r = Runner(planet=planet, planet_name="Earth", partial_kwargs=["planet_name"])
    text_got = r.render("Hello {{planet()}}")
    text_want = "Hello Earth"
    assert text_got == text_want
    file_got = r.render("tests/files/partial_planet.sql")
    file_want = "Hello Earth, Goodbye Earth"
    assert file_got == file_want



import pytest
from absql.files import parse


@pytest.fixture
def simple_sql_path():
    return "tests/files/simple.sql"


@pytest.fixture
def simple_yml_path():
    return "tests/files/simple.yml"


def test_simple_sql(simple_sql_path):
    res = parse(simple_sql_path)
    assert "sql" in res.keys()
    assert "my_table_placeholder" in res.keys()
    assert res["my_table_placeholder"] == "my_table"
    assert res["sql"] == "SELECT * FROM {{my_table_placeholder}}"


def test_simple_yml(simple_yml_path):
    res = parse(simple_yml_path)
    assert "sql" in res.keys()
    assert "my_table_placeholder" in res.keys()
    assert res["my_table_placeholder"] == "my_table"
    assert res["sql"] == "SELECT * FROM {{my_table_placeholder}}"



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ABSQL",
    version="0.3.2",
    author="Chris Cardillo",
    author_email="cfcardillo23@gmail.com",
    description="A rendering engine for templated SQL",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/chriscardillo/ABSQL",
    packages=setuptools.find_packages(),
    install_requires=[
        "colorama",
        "flatdict",
        "Jinja2",
        "pendulum",
        "PyYaml",
        "SQLAlchemy",
        "sql-metadata",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from absql import Runner as r


def test_render_text():
    template = "SELECT * FROM {{my_table_placeholder}}"
    result = r.render_text(template, my_table_placeholder="my_table")
    assert result == "SELECT * FROM my_table"


def test_leave_unknown_vars_alone():
    template = "SELECT * FROM {{my_table_placeholder}} WHERE date = '{{my_date}}'"
    result = r.render_text(template, my_table_placeholder="my_table")
    assert result == "SELECT * FROM my_table WHERE date = '{{ my_date }}'"


# Would be nice to just use jinja for this
# https://stackoverflow.com/questions/71374498/ignore-unknown-functions-in-jinja2
def test_leave_unknown_functions_alone():
    template = "SELECT * FROM {{my_table_placeholder}} WHERE date = '{{ get_date() }}'"
    result = r.render_text(template, my_table_placeholder="my_table", replace_only=True)
    assert result == "SELECT * FROM my_table WHERE date = '{{ get_date() }}'"


def test_nested_replace():
    context = {"params": {"foo": "bar", "biz": "baz"}}
    template = "{{ params.foo }} and {{  params.biz  }}"
    got = r.render_text(template, replace_only=True, **context)
    want = "bar and baz"
    assert got == want


def test_pretty_encode():
    got = r.render_text(
        """SELECT * FROM "my_schema"."my_table" WHERE my_col = 'my_value'""",
        pretty_encode=True,
    )
    want = """\x1b[1m\x1b[96mSELECT\x1b[0m * \x1b[1m\x1b[96mFROM\x1b[0m \x1b[95m"my_schema"\x1b[39m.\x1b[95m"my_table"\x1b[39m \x1b[1m\x1b[96mWHERE\x1b[0m my_col = \x1b[95m\'my_value\'\x1b[39m"""  # noqa
    assert got == want


def test_filters():
    def alter_text(x, upper=True):
        if upper:
            return x.upper()
        return x.lower()

    you = "yOu"

    upper_want = "hey YOU"
    upper_got = r.render_text(
        "hey {{person | case_it}}", person=you, case_it=alter_text
    )
    assert upper_got == upper_want

    lower_want = "hey you"
    lower_got = r.render_text(
        "hey {{person | case_it(upper=False)}}", person=you, case_it=alter_text
    )
    assert lower_want == lower_got

    partial_want = "hey you"
    partial_got = r.render_text(
        "hey {{person | case_it}}",
        person=you,
        case_it=alter_text,
        upper=False,
        partial_kwargs=["upper"],
    )
    assert partial_want == partial_got

    pipeline_want = "hey YOU"
    pipeline_got = r.render_text(
        "hey {{person | case_it(upper=False) | case_it}}",
        person=you,
        case_it=alter_text,
    )
    assert pipeline_want == pipeline_got
